#include <stdlib.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>

#define LED_COUNT 17

#define setbit(x, bit, val) 	(x = (x & ~(1 << (bit))) | ((val)?(1 << (bit)):(0)))

int position;
char direction;
unsigned char state[LED_COUNT];

void init() {
	int i;

	// drive-line setup
	PORTB = 0;
	PORTD = 0;
	PORTA = 0;
	DDRB = 0xff;
	DDRD = 0x7f;
	DDRA = 0x03;

	// clear display state
	position = 0;
	direction = 1;
	for(i = 0; i < LED_COUNT; i++) state[i] = 1 << i;

	// timer for stepping position
	TCCR0A = _BV(WGM01);			// Clear timer on compare match mode
	TCCR0B = _BV(CS02) | _BV(CS00);		// Clock prescale by 1024
	OCR0A = 255;				// About 30 Hz
	TIMSK = _BV(OCIE0A);			// Enable compare match interrupts
	sei();
}

void setLed(char n, char x) {
	if(n < 8) {
		setbit(PORTB, n, x);
	} else if(n < 15) {
		setbit(PORTD, n - 8, x);
	} else {
		setbit(PORTA, n - 15, x);
	}
}

ISR(TIMER0_COMPA_vect) {
	state[position] = 0xff;
	position += direction;
	if(position == 0 || position == LED_COUNT - 1) direction = -direction;
}

void display() {
	int i, j;

	state[position] = 0xff; // you may wish to comment this out...

	for(i = 0; i < LED_COUNT; i++) {
		if(state[i] > 0) setLed(i, 1);
	}

	for(j = 0; j < 256; j += 8) {
		for(i = 0; i < LED_COUNT; i++) {
			if(state[i] < j) setLed(i, 0);
		}
	}
}

void decay() {
	int i;
	for(i = 0; i < LED_COUNT; i++) {
		// pseudo-exponential decay
		if(state[i] > 0) state[i] -= (state[i] >> 3) + 1;
	}
}

int main() {
	unsigned char n = 0;

	init();
	while(1) {
		display();
		decay();
	}

	return 0; // unreached
}
